# This file is a part of thelatinlibrary
#
# Copyright (c) 2018 The thelatinlibrary Authors (see AUTHORS)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dataclasses import dataclass
from typing import List

from .tome import Tome


@dataclass
class Section:
    """
    The Section class represents a
    selection of books written by a
    latin author.
    """

    name: str
    """
    The name of the section.
    """

    tomes: List[Tome]
    """
    The tomes contained into the section.
    """

    def contains(self, link: str) -> bool:
        """
        Used to know if a link is already contained
        into the section.

        Args:
            link (str): The link that is being checked.

        Example:
            You can check if a link is contained into this section by using
            this method.

            >>> import thelatinlibrary
            >>> caesar = thelatinlibrary.get_author("http://www.thelatinlibrary.com/caes.html")
            >>> caesar
            Author(link='http://www.thelatinlibrary.com/caes.html', name='Caesar')
            >>> caesar.works
            Works(sections=[Section(name='COMMENTARIORUM LIBRI VII DE BELLO GALLICO CUM A. HIRTI SUPPLEMENTO',
            ..., tomes=[])
            >>> section = caesar.works.sections[0]
            >>> section.contains("http://www.thelatinlibrary.com/caesar/gall1.shtml")
            True
            >>> section.contains("http://www.thelatinlibrary.com/caesar/bc3.shtml")
            False

        Returns:
            bool: True if the link is found, otherwise False.
        """
        return any(map(lambda tome: link == tome.link, self.tomes))



# This file is a part of thelatinlibrary
#
# Copyright (c) 2018 The thelatinlibrary Authors (see AUTHORS)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dataclasses import dataclass
from typing import List, NamedTuple

import requests
from bs4 import BeautifulSoup

from thelatinlibrary import settings
from .section import Section
from .tome import Tome

_endpoint = "http://www.thelatinlibrary.com/"
_exceptions = ["christian.html", "classics.html", "index.html",
               "medieval.html", "misc.html", "neo.html", "readme.html"]

Works = NamedTuple('Works', [('sections', List[Section]), ('tomes', List[Tome])])


@dataclass
class Author:
    """
    The Author class represents a
    latin author/writer.
    """

    link: str
    """
    The link to the author.
    """

    name: str
    """
    The name of the author.
    """

    _works = Works([], [])
    """
    Used to store the result of the
    works getter.
    """

    @property
    def works(self) -> Works:
        """
        A list of all the books the author
        have written.

        Example:
            You can get what an author wrote using this method.

            >>> import thelatinlibrary
            >>> caesar = thelatinlibrary.get_author("http://www.thelatinlibrary.com/caes.html")
            >>> caesar
            Author(link='http://www.thelatinlibrary.com/caes.html', name='Caesar')
            >>> caesar.works
            Works(sections=[Section(name='COMMENTARIORUM LIBRI VII DE BELLO GALLICO CUM A. HIRTI SUPPLEMENTO',
            ..., tomes=[])
        """
        if settings.save and self._works.count([]) < 2:
            return self._works

        text = requests.get(self.link).text
        parser = BeautifulSoup(text, "html5lib")

        if '404' in parser.title.string:
            works = Works([], [])

            if settings.save_on_error:
                self._works = works

            return works

        sections = []
        tomes = []

        tables = list(parser.find_all("table"))
        titles = list(filter(lambda tag: 'work' not in tag.attrs, parser.find_all("div")))
        work_attributes = parser.find_all("h2", class_="work")
        works = list(filter(lambda tag: 'class' not in tag.attrs and tag.text,
                            parser.find_all("p")))

        if work_attributes:
            div_works = list(parser.find_all("div", class_="work"))

            for work_title in work_attributes:
                work_link = work_title.find("a")
                title = work_title.text.strip().rstrip()

                if work_link and work_link.text.strip().rstrip() == title:
                    tomes.append(Tome(link=_endpoint + work_link['href'],
                                      name=title.strip().rstrip()))
                elif ":" in title:
                    title_links = list(work_title.find_all("a"))

                    if title_links:
                        section = Section(tomes=[], name=title[:title.index(":")].strip().rstrip())

                        for title_link in title_links:
                            if 'href' not in title_link.attrs:
                                continue

                            section.tomes.append(Tome(link=_endpoint + title_link['href'],
                                                      name=title_link.string.strip().rstrip()))

                        sections.append(section)
                else:
                    sections.append(Section(tomes=[], name=" ".join(work_title.strings)))

            for index, section in enumerate(sections):
                try:
                    links = div_works[index].find_all("a")

                    for link in links:
                        if 'href' not in link.attrs:
                            continue

                        section.tomes.append(Tome(link=_endpoint + link['href'], name=link.string.strip().rstrip()))

                    div_works.pop(index)
                except IndexError:
                    continue

            if div_works:
                for div_work in div_works:
                    links = div_work.find_all("a")

                    for link in links:
                        if 'href' not in link.attrs:
                            continue

                        sections[-1].tomes.append(Tome(link=_endpoint + link['href'],
                                                       name=link.string.strip().rstrip()))
        else:
            tables = tables[:-1]

            if len(titles) == len(tables):
                try:
                    for div, table in zip(titles, tables):
                        if not div.text:
                            continue

                        section = Section(tomes=[], name=div.text.strip().rstrip())
                        links = table.find_all("a")

                        for work_link in links:
                            if 'href' not in work_link.attrs:
                                continue

                            section.tomes.append(Tome(link=_endpoint + work_link['href'],
                                                      name=work_link.string.strip().rstrip()))

                        sections.append(section)
                except ValueError:
                    pass
            elif len(tables) in [2, 3] and len(works) in [1, 2]:
                title_table, books_table = tables[0], tables[1]
                title = title_table.text and title_table.text.strip().rstrip()

                if title:
                    section = Section(tomes=[], name=title)
                    links = books_table.find_all("a")

                    for work_link in links:
                        if 'href' not in work_link.attrs:
                            continue

                        section.tomes.append(Tome(link=_endpoint + work_link['href'],
                                                  name=work_link.string.strip().rstrip()))

                    sections.append(section)
            else:
                for index, title in enumerate(titles):
                    link = title.find("a")

                    if link:
                        if 'href' not in link.attrs or not link.string:
                            continue

                        tomes.append(Tome(link=_endpoint + link['href'], name=link.string.strip().rstrip()))
                    else:
                        if not title.string:
                            continue

                        links = works[index].find_all("a")
                        section = Section(tomes=[], name=title.string.strip().rstrip())

                        for book_link in links:
                            if 'href' not in book_link.attrs:
                                continue

                            section.tomes.append(Tome(link=_endpoint + book_link['href'],
                                                      name=book_link.string.strip().rstrip()))

                        sections.append(section)

        for work in works:
            tables = list(work.find_all("table"))

            if len(tables) > 1:
                title_table = tables[0].find_all("td")

                if title_table and len(title_table) == 1:
                    links = work.find_all("a")
                    section = Section(tomes=[], name=tables[0].find("td").text.strip().rstrip())

                    for link in links:
                        if 'href' not in link.attrs:
                            continue

                        if link['href'] in _exceptions:
                            continue

                        if link.text and link.text.strip().rstrip() != section.name:
                            section.tomes.append(Tome(link=_endpoint + link['href'], name=link.text.strip().rstrip()))

                    sections.append(section)

        if not tomes:
            links = parser.find_all("a")

            for link in links:
                if 'href' not in link.attrs:
                    continue

                if link['href'] in _exceptions or any(map(lambda s: s.contains(_endpoint + link['href']), sections)):
                    continue

                if link.string:
                    tomes.append(Tome(link=_endpoint + link['href'], name=link.string.strip().rstrip()))

                font = link.find("font")

                if font:
                    tomes.append(Tome(link=_endpoint + link['href'], name=font.text.strip().rstrip()))

        if not tomes and not sections:
            tomes.append(Tome(link=self.link, name=self.name))

        works = Works(sections, tomes)

        if settings.save:
            self._works = works

        return works



# This file is a part of thelatinlibrary
#
# Copyright (c) 2018 The thelatinlibrary Authors (see AUTHORS)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from html import unescape
from typing import List, Optional

import requests
from bs4 import BeautifulSoup

from thelatinlibrary import settings
from .authors import Author
from .exceptions import NoAuthorException

_authors = []
_endpoint = "http://www.thelatinlibrary.com/"
_exceptions = ["christian.html", "classics.html", "index.html",
               "medieval.html", "misc.html", "neo.html", "readme.html"]


def get_author(url: str) -> Author:
    """
    Used to return an Author by passing
    an URL as an argument.
    
    Example:
        All authors are associated to a link, so it is easy to get an author
        using it.

        >>> import thelatinlibrary
        >>> caesar = thelatinlibrary.get_author("http://www.thelatinlibrary.com/caes.html")
        >>> caesar
        Author(link='http://www.thelatinlibrary.com/caes.html', name='Caesar')

    Args:
      url (str): The thelatinlibrary.com URL.

    Returns:
      Author: The Author object.
    """
    global _authors

    url = url.replace("https://", "http://")

    if not url.startswith(_endpoint):
        url = _endpoint + url

    if settings.save_on_methods:
        try:
            return next(filter(lambda writer: writer.link == url, _authors))
        except StopIteration:
            pass

    text = requests.get(url).text
    parser = BeautifulSoup(text, "html5lib")

    if '404' in parser.title.string:
        raise NoAuthorException("No author has been found on %s" % url)
    else:
        author = Author(link=url, name=parser.title.string.strip().rstrip())

        if settings.save_on_methods and author not in _authors:
            _authors.append(author)

        return author


def get_author_by_name(name: str) -> Author:
    """
    Used to return an Author by passing
    a name as an argument.

    Args:
        name (str): The Author name.

    Example:
        All authors are listed in thelatinlibrary.com index, so it is
        easy to get an author using it.

        >>> import thelatinlibrary
        >>> caesar = thelatinlibrary.get_author_by_name("caesar")
        >>> caesar
        Author(link='http://www.thelatinlibrary.com/caes.html', name='Caesar')

    Raises:
        NoAuthorException: If no author has been found.

    Returns:
        Author: The Author object.
    """
    if settings.save_on_methods:
        try:
            return next(filter(lambda writer: writer.name.lower() == name.lower(), _authors))
        except StopIteration:
            pass

    text = requests.get(_endpoint + "indices.html").text
    parser = BeautifulSoup(text, "html5lib")

    tables = parser.find_all("td")

    for table in tables:
        links = table.find_all("a")

        for link in links:
            if link['href'] not in _exceptions:
                if link.string.strip().rstrip().lower() == name.lower():
                    author = Author(name=unescape(link.string.strip().rstrip()), link=_endpoint + link['href'])

                    if settings.save_on_methods and author not in _authors:
                        _authors.append(author)

                    return author

    raise NoAuthorException("There's no %s in the thelatinlibrary.com index." % name)


def get_authors(reload: Optional[bool] = False) -> List[Author]:
    """
    Used to return a list of all the classic
    authors found on thelatinlibrary.com.

    Args:
        reload (Optional[bool]): If it should be ignored the already stored authors. (Default value = False)

    Example:
        There's thelatinlibrary.com index, the function just returns it.

        >>> import thelatinlibrary
        >>> authors = thelatinlibrary.get_authors()
        >>> authors
        [Author(link='http://www.thelatinlibrary.com/abel.historia.html', name='Abelard'), ...,
        Author(link='http://www.thelatinlibrary.com/waardenburg.html', name='Waardenburg')]


    Raises:
        NoAuthorException: If no author has been found.

    Returns:
        List[Author]: A list of all the class authors.
    """
    global _authors

    if settings.save and _authors and not reload:
        return _authors

    text = requests.get(_endpoint + "indices.html").text
    parser = BeautifulSoup(text, "html.parser")

    authors, books, tomes = [], [], []
    tables = parser.find_all("td")

    for table in tables:
        links = table.find_all("a")

        for link in links:
            if link['href'] not in _exceptions:
                authors.append(Author(name=unescape(link.string.strip().rstrip()), link=_endpoint + link['href']))

    if settings.save:
        _authors = authors

    return authors



# This file is a part of thelatinlibrary
#
# Copyright (c) 2018 The thelatinlibrary Authors (see AUTHORS)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dataclasses import dataclass
from typing import List

import requests
from bs4 import BeautifulSoup

from thelatinlibrary import settings


@dataclass
class Tome:
    """
    The Tome class represents one of the
    multiple volumes used to write a book.
    """

    link: str
    """
    The link to the tome.
    """

    name: str
    """
    The name of the tome.
    """

    _text = ""
    """
    Used to store the result of the
    text getter.
    """

    @property
    def text(self) -> List[str]:
        """
        The text contained in the tome.
        """
        if settings.save and self._text:
            return self._text

        text = requests.get(self.link).text
        parser = BeautifulSoup(text, "html5lib")

        if '404' in parser.title.string:
            return []

        paragraphs = []
        texts = parser.find_all("p")

        for text in texts:
            paragraphs.append(text)

        if settings.save:
            self._text = paragraphs

        return paragraphs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is a part of thelatinlibrary
#
# Copyright (c) 2018 The thelatinlibrary Authors (see AUTHORS)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import io
import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as file:
        long_description = '\n' + file.read()
except FileNotFoundError:
    long_description = 'An unofficial thelatinlibrary.com client'

setup(
    author='hearot',
    author_email='gabriel@hearot.it',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    description='An unofficial thelatinlibrary.com client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='GNU GENERAL PUBLIC LICENSE',
    install_requires=['BeautifulSoup4', 'dataclasses', 'html5lib', 'requests'],
    name='thelatinlibrary',
    packages=['thelatinlibrary'],
    python_requires='>=3.6.0',
    url='https://github.com/hearot/thelatinlibrary',
    version='0.2.0'
)
